/*#################################################################################################
##  |\   ======   /|  ##     Copyright (C) 2013-2014  Maurizio Sartori     ##  |\   ======   /|  ##
##  |__\========/__|  ##  ===============================================  ##  |__\========/__|  ##
##  |\            /|  ##     Written by:  Maurizio Sartori                 ##  |\            /|  ##
##  |  \        /  |  ##     e-mail:      masar@MasarLabs.com              ##  |  \        /  |  ##
##  |    \    /    |  ##     Site:        http://www.masarlabs.com         ##  |    \    /    |  ##
##  |    /=\/=\    |  ##  ===============================================  ##  |    /=\/=\    |  ##
##  |  /========\  |  ##     License GPLv3+: GNU GPL version 3 or later    ##  |  /========\  |  ##
##  |/   ======   \|  ##          http://gnu.org/licenses/gpl.html         ##  |/   ======   \|  ##
###################################################################################################
## $Id: network.c 4108 2014-02-06 17:10:52Z masar@masar $
#################################################################################################*/

static char Psz_SvnId[] __attribute__ ((unused)) =
    "$Id: network.c 4108 2014-02-06 17:10:52Z masar@masar $";

#include "network.h"

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

/*************************************************************************************************/
/** setSocketNonBlocking.
*** Returns -1 on error.
**************************************************************************************************/
int setSocketNonBlocking (int fdSocket)
   {
   int iFlags = fcntl(fdSocket, F_GETFL, 0);
   if (iFlags == -1) {
      return -1;
      }

   return fcntl(fdSocket, F_SETFL, iFlags|O_NONBLOCK);
   }

/*************************************************************************************************/
/** printSocketAddress.
*** Returns -1 on error.
**************************************************************************************************/
const char const * printSocketAddress (struct sockaddr * psa, ssize_t iSize)
   {
   static char achString[NI_MAXHOST + NI_MAXSERV + 2];
   char achHost[NI_MAXHOST];
   char achPort[NI_MAXSERV];
   if (!getnameinfo(psa, iSize, achHost, sizeof(achHost), achPort, sizeof(achPort),
                    NI_NUMERICHOST | NI_NUMERICSERV)) {
      snprintf(achString, sizeof(achString), "%s %s", achHost, achPort);
      return achString;
      }
   else {
      return "error";
      }
   }

/*************************************************************************************************/
/** openTcpServerSocket.
*** Returns -1 on error.
**************************************************************************************************/
int openTcpServerSocket (const char * pchHost, const char * pchPort)
   {
   struct addrinfo aiHints;
   memset(&aiHints, 0, sizeof(aiHints));

   aiHints.ai_family = AF_UNSPEC;         /* Return IPv4 and IPv6 choices */
   aiHints.ai_protocol = IPPROTO_TCP;     /* We want a TCP socket */
   aiHints.ai_socktype = SOCK_STREAM;     /* We want a TCP socket */
   aiHints.ai_flags = AI_PASSIVE;         /* All interfaces */

   struct addrinfo * paiAddresses = NULL;
   int iRc = getaddrinfo(pchHost, pchPort, &aiHints, &paiAddresses);
   if (iRc != 0) {
      syslog(LOG_DEBUG, "getaddrinfo: %s", gai_strerror(iRc));
      return -1;
      }

   int fdSocket = -1;
   for (struct addrinfo * pai = paiAddresses; pai != NULL; pai = pai->ai_next) {
      fdSocket = socket(pai->ai_family, pai->ai_socktype, pai->ai_protocol);
      if (fdSocket == -1) {
         continue;
         }

      int iOptVal = 1;
      if (!setsockopt(fdSocket, SOL_SOCKET, SO_REUSEADDR, &iOptVal, sizeof(iOptVal)) &&
          !bind(fdSocket, pai->ai_addr, pai->ai_addrlen)) {
         syslog(LOG_NOTICE, "Server listening on: %s",
                            printSocketAddress(pai->ai_addr, pai->ai_addrlen));
         break;
         }
      close(fdSocket);
      fdSocket = -1;
      }
   freeaddrinfo(paiAddresses);

   if (fdSocket == -1) {
      return -1;
      }

   if (setSocketNonBlocking(fdSocket)) {
      close(fdSocket);
      syslog(LOG_ERR, "Could not set non blocking");
      return -1;
      }

   if (listen(fdSocket, SOMAXCONN)) {
      syslog(LOG_ERR, "listen error");
      close(fdSocket);
      return -1;
      }
   return fdSocket;
   }

/*************************************************************************************************/
/** openTcpClientSocket.
*** Returns -1 on error.
**************************************************************************************************/
int openTcpClientSocket (const char * pchHost, const char * pchPort)
   {
   struct addrinfo aiHints;
   memset(&aiHints, 0, sizeof(aiHints));

   aiHints.ai_family = AF_UNSPEC;         /* Return IPv4 and IPv6 choices */
   aiHints.ai_protocol = IPPROTO_TCP;     /* We want a TCP socket */
   aiHints.ai_socktype = SOCK_STREAM;     /* We want a TCP socket */
   aiHints.ai_flags = AI_PASSIVE;         /* All interfaces */

   struct addrinfo * paiAddresses = NULL;
   int iRc = getaddrinfo(pchHost, pchPort, &aiHints, &paiAddresses);
   if (iRc != 0) {
      syslog(LOG_DEBUG, "getaddrinfo: %s", gai_strerror(iRc));
      return -1;
      }

   int fdSocket = -1;
   for (struct addrinfo * pai = paiAddresses; pai != NULL; pai = pai->ai_next) {
      fdSocket = socket(pai->ai_family, pai->ai_socktype, pai->ai_protocol);
      if (fdSocket == -1) {
         continue;
         }

      if (!connect(fdSocket, pai->ai_addr, pai->ai_addrlen)) {
         syslog(LOG_DEBUG, "Client connected to: %s",
                printSocketAddress(pai->ai_addr, pai->ai_addrlen));
         break;
         }
      close(fdSocket);
      fdSocket = -1;
      }
   freeaddrinfo(paiAddresses);

   return fdSocket;
   }

/*************************************************************************************************/
/** Writes all the data.
**************************************************************************************************/
int writeAll (int fdSocket, const void *p, size_t iCount)
   {
   while (iCount) {
      int iWritten = send(fdSocket, (uint8_t *)p, iCount, 0);
      if (iWritten <= 0) {
         if (errno != EINTR && errno != EAGAIN) {
            return -1;
            }
         continue;
         }
      iCount -= iCount;
      p = (uint8_t*) p + iWritten;
      }
   return 0;
   }

/*************************************************************************************************/
/** Reads all the data.
**************************************************************************************************/
int readAll (int fdSocket, const void *p, size_t iCount)
   {
   while (iCount) {
      int iRead = recv(fdSocket, (uint8_t*)p, iCount, 0);
      if (iRead == 0) {
         return -1;  /* EOF */
         }
      if (iRead < 0) {
         if (errno != EINTR && errno != EAGAIN ) {
            return -1;
            }
         continue;
         }
      iCount -= iRead;
      p = (uint8_t*) p + iRead;
      }
   return 0;
   }

/*************************************************************************************************/
/** Calculate a partial Checksum.
**************************************************************************************************/
static uint32_t icmpChecksumAdd (const uint8_t * pBuffer, size_t iLen, uint32_t iSum)
   {
   const uint16_t * p = (const uint16_t *) pBuffer;
   while (iLen > 1) {
      iSum += *p;
      p += 1;
      iLen -= 2;
      }

   if (iLen) {
      iSum += *(uint8_t*) p;
      }
   return iSum;
   }

/*************************************************************************************************/
/** Calculate an IPv4 CheckSum.
**************************************************************************************************/
uint16_t icmpChecksum (const uint8_t * pBuffer, size_t iLen)
   {
   uint32_t iSum = icmpChecksumAdd(pBuffer, iLen, 0);

   iSum = (iSum >> 16) + (iSum & 0xFFFF);
   iSum += (iSum >> 16);
   return (uint16_t) ~iSum;
   }

#if 0
// Build IPv6 ICMP pseudo-header and call checksum function (Section 8.1 of RFC 2460).
uint16_t
icmp6_checksum (struct ip6_hdr iphdr, struct icmp6_hdr icmp6hdr, uint8_t *payload, int payloadlen)
{
  char buf[IP_MAXPACKET];
  char *ptr;
  int chksumlen = 0;
  int i;

  ptr = &buf[0];  // ptr points to beginning of buffer buf

  // Copy source IP address into buf (128 bits)
  memcpy (ptr, &iphdr.ip6_src.s6_addr, sizeof (iphdr.ip6_src.s6_addr));
  ptr += sizeof (iphdr.ip6_src);
  chksumlen += sizeof (iphdr.ip6_src);

  // Copy destination IP address into buf (128 bits)
  memcpy (ptr, &iphdr.ip6_dst.s6_addr, sizeof (iphdr.ip6_dst.s6_addr));
  ptr += sizeof (iphdr.ip6_dst.s6_addr);
  chksumlen += sizeof (iphdr.ip6_dst.s6_addr);

  // Copy Upper Layer Packet length into buf (32 bits).
  // Should not be greater than 65535 (i.e., 2 bytes).
  *ptr = 0; ptr++;
  *ptr = 0; ptr++;
  *ptr = (ICMP_HDRLEN + payloadlen) / 256;
  ptr++;
  *ptr = (ICMP_HDRLEN + payloadlen) % 256;
  ptr++;
  chksumlen += 4;

  // Copy zero field to buf (24 bits)
  *ptr = 0; ptr++;
  *ptr = 0; ptr++;
  *ptr = 0; ptr++;
  chksumlen += 3;

  // Copy next header field to buf (8 bits)
  memcpy (ptr, &iphdr.ip6_nxt, sizeof (iphdr.ip6_nxt));
  ptr += sizeof (iphdr.ip6_nxt);
  chksumlen += sizeof (iphdr.ip6_nxt);

  // Copy ICMPv6 type to buf (8 bits)
  memcpy (ptr, &icmp6hdr.icmp6_type, sizeof (icmp6hdr.icmp6_type));
  ptr += sizeof (icmp6hdr.icmp6_type);
  chksumlen += sizeof (icmp6hdr.icmp6_type);

  // Copy ICMPv6 code to buf (8 bits)
  memcpy (ptr, &icmp6hdr.icmp6_code, sizeof (icmp6hdr.icmp6_code));
  ptr += sizeof (icmp6hdr.icmp6_code);
  chksumlen += sizeof (icmp6hdr.icmp6_code);

  // Copy ICMPv6 ID to buf (16 bits)
  memcpy (ptr, &icmp6hdr.icmp6_id, sizeof (icmp6hdr.icmp6_id));
  ptr += sizeof (icmp6hdr.icmp6_id);
  chksumlen += sizeof (icmp6hdr.icmp6_id);

  // Copy ICMPv6 sequence number to buff (16 bits)
  memcpy (ptr, &icmp6hdr.icmp6_seq, sizeof (icmp6hdr.icmp6_seq));
  ptr += sizeof (icmp6hdr.icmp6_seq);
  chksumlen += sizeof (icmp6hdr.icmp6_seq);

  // Copy ICMPv6 checksum to buf (16 bits)
  // Zero, since we don't know it yet.
  *ptr = 0; ptr++;
  *ptr = 0; ptr++;
  chksumlen += 2;

  // Copy ICMPv6 payload to buf
  memcpy (ptr, payload, payloadlen * sizeof (uint8_t));
  ptr += payloadlen;
  chksumlen += payloadlen;

  // Pad to the next 16-bit boundary
  for (i=0; i<payloadlen%2; i++, ptr++) {
    *ptr = 0;
    ptr += 1;
    chksumlen += 1;
  }

  return checksum ((uint16_t *) buf, chksumlen);
}
#endif


/*************************************************************************************************/
/*************************************************************************************************/
