/*#################################################################################################
##  |\   ======   /|  ##     Copyright (C) 2013-2014  Maurizio Sartori     ##  |\   ======   /|  ##
##  |__\========/__|  ##  ===============================================  ##  |__\========/__|  ##
##  |\            /|  ##     Written by:  Maurizio Sartori                 ##  |\            /|  ##
##  |  \        /  |  ##     e-mail:      masar@MasarLabs.com              ##  |  \        /  |  ##
##  |    \    /    |  ##     Site:        http://www.masarlabs.com         ##  |    \    /    |  ##
##  |    /=\/=\    |  ##  ===============================================  ##  |    /=\/=\    |  ##
##  |  /========\  |  ##     License GPLv3+: GNU GPL version 3 or later    ##  |  /========\  |  ##
##  |/   ======   \|  ##          http://gnu.org/licenses/gpl.html         ##  |/   ======   \|  ##
###################################################################################################
## $Id: util.c 4158 2014-02-18 15:21:44Z masar@masar $
#################################################################################################*/

static char Psz_SvnId[] __attribute__ ((unused)) =
    "$Id: util.c 4158 2014-02-18 15:21:44Z masar@masar $";

#include "util.h"

#include <stdio.h>
#include <sys/file.h>

//************************************************************************************************/
//* Data
//************************************************************************************************/
#define PIDFILE_BUFFER  30

volatile sig_atomic_t sig_SignalFlag;

/*************************************************************************************************/
/** Set a signal handler with automatic long function restart
**************************************************************************************************/
int setSignal (int iSignal, __sighandler_t pfnFunction)
   {
   struct sigaction sigAction;

   sigAction.sa_handler = pfnFunction;
   sigemptyset (&sigAction.sa_mask);
   sigAction.sa_flags = ((iSignal == SIGALRM) ? SA_INTERRUPT : SA_RESTART);

   return sigaction(iSignal, &sigAction, NULL);
   }

/*************************************************************************************************/
/** Simple signal handler, increment sig_SignalFlag.
**************************************************************************************************/
void signalHandlerIncrement (int iSignal)
   {
   (void) iSignal;
   sig_SignalFlag += 1;
   }

/*************************************************************************************************/
/** Compares two timespec.
**************************************************************************************************/
long tsCompare (const struct timespec * pTime1, const struct timespec * pTime2)
   {
   if (pTime1->tv_sec == pTime2->tv_sec) {
      return pTime1->tv_nsec - pTime2->tv_nsec;
      }
   return pTime1->tv_sec - pTime2->tv_sec;
   }

/*************************************************************************************************/
/** Delta two timespec: pResult = abs(pTime1 - pTime2) = abs(pTime2 - pTime1)
**************************************************************************************************/
void tsDelta (const struct timespec * pTime1, const struct timespec * pTime2,
              struct timespec * pResult)
   {
   pResult->tv_sec = pTime1->tv_sec - pTime2->tv_sec;
   pResult->tv_nsec = pTime1->tv_nsec - pTime2->tv_nsec;
   if (pResult->tv_nsec < 0) {
      pResult->tv_sec -= 1;
      pResult->tv_nsec += 1000000000L;
      }
   if (pResult->tv_sec < 0) {
      pResult->tv_sec = -pResult->tv_sec - 1;
      pResult->tv_nsec = 1000000000L - pResult->tv_nsec;
      }
   }

/*************************************************************************************************/
/** Adds two timespec: *pResult = *pTime1 + *pTime2
**************************************************************************************************/
void tsAdd (const struct timespec * pTime1, const struct timespec * pTime2,
            struct timespec * pResult)
   {
   pResult->tv_sec = pTime1->tv_sec + pTime2->tv_sec;
   pResult->tv_nsec = pTime1->tv_nsec + pTime2->tv_nsec;
   if (pResult->tv_nsec >= 1000000000L) {
      pResult->tv_sec += 1;
      pResult->tv_nsec -= 1000000000L;
      }
   }

/*************************************************************************************************/
/** Write the pidfile and returns the file descriptor.
**************************************************************************************************/
int writePidfile (const char * pFilename)
   {
   int iFd = open(pFilename, O_RDWR | O_CREAT | O_CLOEXEC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
   if (iFd < 0) {
      return iFd;
      }

   if (flock(iFd, LOCK_EX | LOCK_NB) == -1) {
      close(iFd);
      return -1;
      }

   char achBuffer[PIDFILE_BUFFER];
   int iLen = snprintf(achBuffer, PIDFILE_BUFFER, "%ld\n", (long) getpid());
   if ((iLen >= PIDFILE_BUFFER) ||
       (ftruncate(iFd, 0) < 0) ||
       (write(iFd, achBuffer, iLen + 1) != iLen + 1) ||
       (fsync(iFd) < 0)) {
      close(iFd);
      return -1;
      }

   return iFd;
   }

/*************************************************************************************************/
/** Hex Dumps.
*** Usage:
***    ssize_t iOffset = 0;
***    const char const * pStr = hexDump(pData, iLen, &iOffset);
***    while (pStr != NULL) {
***       puts(pStr);
***       pStr = hexDump(pData, iLen, &iOffset);
***       }
**************************************************************************************************/
const char const * hexDump (const void * pData, ssize_t iLen, ssize_t * piOffset)
   {
   static char achOut[81];
   static char achHex[16] = { '0', '1', '2', '3', '4', '5', '6', '7',
                              '8', '9', 'A', 'B', 'C', 'F', 'E', 'F' };

   int iInx = *piOffset;
   if (iLen <= iInx) {
      return NULL;
      }

   const char * pIn = (const char *) pData + iInx;
   char * pAsc = achOut + 54;
   char * pHex = achOut;
   *pHex++ = achHex[(iInx >> 12) & 0xF];
   *pHex++ = achHex[(iInx >>  8) & 0xF];
   *pHex++ = achHex[(iInx >>  4) & 0xF];
   *pHex++ = achHex[(iInx >>  0) & 0xF];
   *pHex++ = ':';

   for (int i = 0; i < 16; i += 1) {
      *pHex++ = ' ';
      if (iInx < iLen) {
         *pHex++ = achHex[(*pIn >>  4) & 0xF];
         *pHex++ = achHex[(*pIn >>  0) & 0xF];
         *pAsc++ = ((*pIn >= ' ') && (*pIn <= '~')) ? *pIn : '.';
         pIn += 1;
         iInx += 1;
         }
      else {
         *pHex++ = ' ';
         *pHex++ = ' ';
         *pAsc++ = '.';
         }
      }
   *pHex++ = ' ';
   *pAsc++ = 0;
   *piOffset = iInx;
   return achOut;
   }

/*************************************************************************************************/
/*************************************************************************************************/
